/*
 * Copyright (c) 2005-2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef FAKE_NEOBOTIX
#include <stdio.h>
#endif

#include "jlocoPrivate.h"
#include "libmat.h"
#include <math.h>


static JLOCO_PRIVATE private;


JLOCO_PRIVATE *jlocoPrivate = &private;


void initOdometry(JLOCO_PRIVATE * jp,
		  double x,
		  double y,
		  double theta,
		  double vxx,
		  double vyy,
		  double vtt,
		  double cntrl_dist)
{
  const double st = sin(theta);
  const double ct = cos(theta);
  
  jp->odo.robotMisc.sinTheta = st;
  jp->odo.robotMisc.cosTheta = ct;
  // these fields keep their old values: s, cR, dTheta, dS
  
  // the person who came up with the "x vs xC" naming scheme should be
  // slapped in the face...
  // ATTENTION:
  //   xC stands for the CENTER coordinate, which is usually denoted x, but
  //   x stands for the CONTROL coordinate
  //   the same holds for y and yC
  jp->odo.robot.xRob      = x;
  jp->odo.robot.yRob      = y;
  jp->odo.robot.theta     = theta;
  jp->odo.robot.xRef      = x + ct * cntrl_dist;
  jp->odo.robot.yRef      = y + st * cntrl_dist;
  // these fields keep their old values: v, w
  
  dms_fill(jp->odo.odoError.var,     3, 3, vxx, 0, 0, vyy, 0, vtt);
  
  // don't touch jp->odo.{right,left}Wheel
}


//////////////////////////////////////////////////
// begin
// FAKE_NEOBOTIX implementation

#ifdef FAKE_NEOBOTIX

#ifndef absval
# define absval(x) ( (x) > 0 ? (x) : -(x) )
#endif


static NEO_STR fake_neo;

NEO_STR *neoInit(char *dev)
{
  PVDEBUG("FAKE neoInit(%s)\n", dev);
  fake_neo.lvel = 0;
  fake_neo.rvel = 0;
  fake_neo.gphi = 0;
  return &fake_neo;
}


void neoEnd(NEO_STR *neo)
{
  PVDEBUG("FAKE neoEnd()\n");
}


STATUS canMotorsStart(NEO_STR *neo)
{
  PVDEBUG("FAKE canMotorsStart()\n");
  return OK;
}


STATUS canMotorsReset(NEO_STR *neo)
{
  PVDEBUG("FAKE canMotorsReset()\n");
  return OK;
}


CAN_MOTOR_STATE canMotorsState(NEO_STR *neo, CAN_MOTOR_ID id)
{
  PVDEBUG("FAKE canMotorsState()\n");
  return ST_OPERATION_ENABLED;
}


void canMotorsGetGearDeltaPosVelRadS(NEO_STR *neo,
				     double *lDeltaPos, double *lvel,
				     double *rDeltaPos, double *rvel)
{
  PVDEBUG("FAKE\n");
  *lDeltaPos = neo->lvel * jlocoPrivate->cmdTaskPeriod;
  *lvel      = neo->lvel;
  *rDeltaPos = neo->rvel * jlocoPrivate->cmdTaskPeriod;
  *rvel      = neo->rvel;
}


void canMotorsSetVelGearRadS(NEO_STR *neo, double lvel, double rvel)
{
  static double lvel_prev = 0;
  static double rvel_prev = 0;
  if((absval(lvel - lvel_prev) > 1e-3) || (absval(rvel - rvel_prev) > 1e-3))
    PVDEBUG("FAKE canMotorsSetVelGearRadS(%f, %f)\n", lvel, rvel);
  lvel_prev = lvel;
  rvel_prev = rvel;
  neo->lvel = lvel;
  neo->rvel = rvel;
}


void canGyroRequestConnect(NEO_STR *neo)
{
  PVDEBUG("FAKE canGyroRequestConnect()\n");
}


void canGyroZero(NEO_STR *neo)
{
  PVDEBUG("FAKE canGyroZero()\n");
  neo->gphi = 0;
}


void canGyroRequestAngle(NEO_STR *neo)
{
  double sd, thetad, qdl, qdr;
  PVDEBUG("FAKE\n");
  
#if INVERT_LEFT_WHEEL == 1
  qdl = - neo->lvel;
#elif INVERT_LEFT_WHEEL == 0
  qdl = neo->lvel;
#else
# error INVERT_LEFT_WHEEL must be 0 or 1
#endif
#if INVERT_RIGHT_WHEEL == 1
  qdr = - neo->rvel;
#elif INVERT_RIGHT_WHEEL == 0
  qdr = neo->rvel;
#else
# error INVERT_RIGHT_WHEEL must be 0 or 1
#endif
  speedWheelsToGlob(qdr, qdl, &sd, &thetad,
		    jlocoPrivate->motRr_cor, jlocoPrivate->motRl_cor,
		    jlocoPrivate->motW_cor);
  neo->gphi += thetad * jlocoPrivate->cmdTaskPeriod;
}

double canGyroGetAngle(NEO_STR *neo)
{
  PVDEBUG("FAKE\n");
  return neo->gphi;
}

void  canIORequestConnect(NEO_STR *neo)
{
  PVDEBUG("FAKE\n");
}

int canIOGetVBatt(NEO_STR *neo)
{
  PVDEBUG("FAKE\n");
  return JLOCO_MIN_BATTERY_LEVEL;
}

unsigned short canIOGetDigiIn(NEO_STR *neo)
{
  PVDEBUG("FAKE\n");
  return 0;
}

void canIORequestVBatt(NEO_STR *neo)
{
  PVDEBUG("FAKE\n");
}

void canIORequestDigiIn(NEO_STR *neo)
{
  PVDEBUG("FAKE\n");
}

#endif // FAKE_NEOBOTIX

// end
// FAKE_NEOBOTIX implementation
//////////////////////////////////////////////////
